<?php namespace Myriad\Queue\Drivers\Google;

use Carbon\Carbon;
use Google\Cloud\PubSub\Message;
use Google\Cloud\PubSub\Subscription;
use Illuminate\Container\Container;
use Illuminate\Contracts\Queue\Job as JobContract;
use Illuminate\Queue\Jobs\Job;

class PubSubQueueJob extends Job implements JobContract {

  /**
   * @var PubSubQueue The queue
   */
  private $pubSubQueue;

  /**
   * @var Subscription The pub/sub subscription
   */
  private $subscription;

  /**
   * @var Message The message with encrypted payload
   */
  private $message;

  /**
   * @var array The message payload
   */
  private $payload;

  /**
   * PubSubQueueJob constructor.
   *
   * @param PubSubQueue $pubSubQueue
   * @param Container $container
   * @param Subscription $subscription
   * @param $queue
   * @param Message $message
   * @param array $payload
   */
  public function __construct(PubSubQueue $pubSubQueue, Container $container, Subscription $subscription, $queue, Message $message, array $payload) {
    $this->pubSubQueue = $pubSubQueue;
    $this->container = $container;
    $this->subscription = $subscription;
    $this->queue = $queue;
    $this->message = $message;
    $this->payload = $payload;
  }

  /**
   * The IoC container instance.
   *
   * @var \Illuminate\Container\Container
   */
  protected $container;

  /**
   * Indicates if the job has been deleted.
   *
   * @var bool
   */
  protected $deleted = false;

  /**
   * Indicates if the job has been released.
   *
   * @var bool
   */
  protected $released = false;

  /**
   * Indicates if the job has failed.
   *
   * @var bool
   */
  protected $failed = false;

  /**
   * The name of the queue the job belongs to.
   *
   * @var string
   */
  protected $queue;

  /**
   * Fire the job.
   *
   * @return void
   */
  public function fire() {
    $payload = $this->payload();

    list($class, $method) = CompatJobName::parse($payload['job']);

    $this->instance = $this->resolve($class);
    $this->instance->{$method}($this, $payload['data']);
  }

  /**
   * Renew the modify ack deadline for another n seconds.
   *
   * @param $seconds
   */
  public function renew($seconds) {
    if ($seconds > PubSubQueue::MAX_ACK_SECONDS) {
      throw new \RuntimeException('Max renew time for Google Pub/Sub is ' . PubSubQueue::MAX_ACK_SECONDS . ' seconds.');
    }
    $this->subscription->modifyAckDeadline($this->message, $seconds);
  }

  /**
   * Delete the job from the queue.
   *
   * @return void
   */
  public function delete() {
    $this->deleted = true;
    $this->subscription->acknowledge($this->message);
  }

  /**
   * Determine if the job has been deleted.
   *
   * @return bool
   */
  public function isDeleted() {
    return $this->deleted;
  }

  /**
   * Release the job back into the queue.
   *
   * @param  int $delay
   * @return void
   */
  public function release($delay = 0) {
    $this->released = true;
    $this->pubSubQueue->pushRaw($this->pubSubQueue->constructPayload($this->payload(), null, ['availableAt' => Carbon::now()->addSecond($delay)->timestamp, 'attempts' => $this->attempts() + 1]), $this->queue);
    $this->subscription->acknowledge($this->message);
  }

  /**
   * Determine if the job was released back into the queue.
   *
   * @return bool
   */
  public function isReleased() {
    return $this->released;
  }

  /**
   * Determine if the job has been deleted or released.
   *
   * @return bool
   */
  public function isDeletedOrReleased() {
    return $this->isDeleted() || $this->isReleased();
  }

  /**
   * Determine if the job has been marked as a failure.
   *
   * @return bool
   */
  public function hasFailed() {
    return $this->failed;
  }

  /**
   * Mark the job as "failed".
   *
   * @return void
   */
  public function markAsFailed() {
    $this->failed = true;
  }

  /**
   * Process an exception that caused the job to fail.
   *
   * @param  \Exception $e
   * @return void
   */
  public function failed($e = null) {
    $this->markAsFailed();

    $payload = $this->payload();

    list($class, $method) = CompatJobName::parse($payload['job']);

    if (method_exists($this->instance = $this->resolve($class), 'failed')) {
      $this->instance->failed($payload['data'], $e);
    }
  }

  /**
   * Resolve the given class.
   *
   * @param  string $class
   * @return mixed
   */
  protected function resolve($class) {
    return $this->container->make($class);
  }

  /**
   * Get the decoded body of the job.
   *
   * @return array
   */
  public function payload() {
    return $this->payload;
  }

  /**
   * The number of times to attempt a job.
   *
   * @return int|null
   */
  public function maxTries() {
    return array_get($this->payload(), 'maxTries');
  }

  /**
   * The number of seconds the job can run.
   *
   * @return int|null
   */
  public function timeout() {
    return array_get($this->payload(), 'timeout');
  }

  /**
   * Get the name of the queued job class.
   *
   * @return string
   */
  public function getName() {
    return $this->payload()['job'];
  }

  /**
   * Get the resolved name of the queued job class.
   *
   * Resolves the name of "wrapped" jobs such as class-based handlers.
   *
   * @return string
   */
  public function resolveName() {
    return CompatJobName::resolve($this->getName(), $this->payload());
  }

  /**
   * Get the name of the queue the job belongs to.
   *
   * @return string
   */
  public function getQueue() {
    return $this->queue;
  }

  /**
   * Get the service container instance.
   *
   * @return \Illuminate\Container\Container
   */
  public function getContainer() {
    return $this->container;
  }

  /**
   * Get the number of times the job has been attempted.
   *
   * @return int
   */
  public function attempts() {
    return array_get($this->payload(), 'attempts', 1);
  }

  /**
   * Get the raw body string for the job.
   *
   * @return array
   */
  public function getRawBody() {
    return json_encode($this->payload);
  }

  /**
   * Get the job's identifier
   *
   * @return array|mixed|string|null
   */
  public function getJobId(){
    return data_get($this->payload(), 'id');
  }
}